@@ -56,6 +56,8 @@
 //!
 //! ```
 #![cfg_attr(test, feature(no_coverage))]
+#![macro_use]
+extern crate zstd;
 
 use std::collections::BTreeMap;
 use std::fs;
@@ -67,6 +69,7 @@ use std::sync::{
 };
 
 const SSTABLE_DIR: &str = "sstables";
+const U64_SZ: usize = std::mem::size_of::<u64>();
 
 #[derive(Debug, Clone, Copy)]
 #[cfg_attr(
@@ -80,7 +83,7 @@ pub struct Config {
     /// multiple versions of most of the database's keys exist
     /// in multiple sstables, but should never happen for workloads
     /// where mostly new keys are being written.
-    pub max_space_amp: u64,
+    pub max_space_amp: u8,
     /// When the log file exceeds this size, a new compressed
     /// and compacted sstable will be flushed to disk and the
     /// log file will be truncated.
@@ -89,15 +92,15 @@ pub struct Config {
     /// ranges of sstables to merge, it will require all sstables
     /// to be at least 1/`merge_ratio` * the size of the first sstable
     /// in the contiguous window under consideration.
-    pub merge_ratio: u64,
+    pub merge_ratio: u8,
     /// When the background compactor thread looks for ranges of
     /// sstables to merge, it will require ranges to be at least
     /// this long.
-    pub merge_window: usize,
+    pub merge_window: u8,
     /// All inserts go directly to a `BufWriter` wrapping the log
     /// file. This option determines how large that in-memory buffer
     /// is.
-    pub log_bufwriter_size: usize,
+    pub log_bufwriter_size: u32,
     /// The level of compression to use for the sstables with zstd.
     pub zstd_sstable_compression_level: u8,
 }
@@ -219,7 +222,7 @@ impl<const K: usize, const V: usize> Worker<K, V> {
 
         log::debug!("disk size: {} mem size: {}", on_disk_size, self.db_sz);
         if self.sstable_directory.len() > 1
-            && on_disk_size / (self.db_sz + 1) > self.config.max_space_amp
+            && on_disk_size / (self.db_sz + 1) > self.config.max_space_amp as u64
         {
             log::debug!(
                 "performing full compaction, decompressed on-disk \
@@ -232,7 +235,7 @@ impl<const K: usize, const V: usize> Worker<K, V> {
             return Ok(());
         }
 
-        if self.sstable_directory.len() < self.config.merge_window {
+        if self.sstable_directory.len() < self.config.merge_window as usize {
             return Ok(());
         }
 
@@ -240,12 +243,12 @@ impl<const K: usize, const V: usize> Worker<K, V> {
             .sstable_directory
             .iter()
             .collect::<Vec<_>>()
-            .windows(self.config.merge_window)
+            .windows(self.config.merge_window as usize)
         {
             if window
                 .iter()
                 .skip(1)
-                .all(|w| *w.1 * self.config.merge_ratio > *window[0].1)
+                .all(|w| *w.1 * self.config.merge_ratio as u64 > *window[0].1)
             {
                 let run_to_compact: Vec<u64> = window.into_iter().map(|(id, _sum)| **id).collect();
 
@@ -365,10 +368,13 @@ fn write_sstable<const K: usize, const V: usize>(
         .write(true)
         .open(&sst_path)?;
 
-    let mut bw = BufWriter::new(
-        zstd::Encoder::new(file, config.zstd_sstable_compression_level as _)
-            .expect("zstd encoder failure"),
-    );
+    let max_zstd_level = zstd::compression_level_range();
+    let zstd_level = config
+        .zstd_sstable_compression_level
+        .min(*max_zstd_level.end() as u8);
+
+    let mut bw =
+        BufWriter::new(zstd::Encoder::new(file, zstd_level as _).expect("zstd encoder failure"));
 
     bw.write_all(&(items.len() as u64).to_le_bytes())?;
 
@@ -534,12 +540,6 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
             }
         }
 
-        let log = fs::OpenOptions::new()
-            .create(true)
-            .read(true)
-            .write(true)
-            .open(path.join("log"))?;
-
         let sstable_directory = list_sstables(path, true)?;
 
         let mut db = BTreeMap::new();
@@ -555,29 +555,27 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 
         let max_sstable_id = sstable_directory.keys().next_back().copied();
 
+        let log = fs::OpenOptions::new()
+            .create(true)
+            .read(true)
+            .write(true)
+            .open(path.join("log"))?;
+
         let mut reader = BufReader::new(log);
 
-        let min_tuple_sz = std::mem::size_of::<u64>().max(K + V);
-        let mut buf = vec![0; 5 + min_tuple_sz];
+        let tuple_sz = U64_SZ.max(K + V);
+        let header_sz = 5;
+        let header_tuple_sz = header_sz + tuple_sz;
+        let mut buf = vec![0; header_tuple_sz];
 
         let mut memtable = BTreeMap::new();
         let mut recovered = 0;
-        let mut write_batch = None;
-        let mut wb_remaining = 0;
-        let mut wb_recovered = 0;
+
+        // write_batch is the pending memtable updates, the number
+        // of remaining items in the write batch, and the number of
+        // bytes that have been recovered in the write batch.
+        let mut write_batch: Option<(_, usize, u64)> = None;
         while let Ok(()) = reader.read_exact(&mut buf) {
-            macro_rules! rewind {
-                () => {
-                    // need to back up a few bytes to chop off the torn log
-                    log::warn!("detected log tear. chopping length to {}", recovered);
-                    let log_file = reader.get_mut();
-                    log_file.seek(io::SeekFrom::Start(recovered))?;
-                    log_file.set_len(recovered)?;
-                    log_file.sync_all()?;
-                    fs::File::open(path.join(SSTABLE_DIR))?.sync_all()?;
-                    break;
-                };
-            }
             let crc_expected: u32 = u32::from_le_bytes(buf[0..4].try_into().unwrap());
             let d: bool = match buf[4] {
                 0 => false,
@@ -586,6 +584,13 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
                     // begin batch
                     let batch_sz_buf: [u8; 8] = buf[5..5 + 8].try_into().unwrap();
                     let batch_sz: u64 = u64::from_le_bytes(batch_sz_buf);
+                    log::debug!("processing batch of len {}", batch_sz);
+
+                    let crc_actual = hash_batch_len(usize::try_from(batch_sz).unwrap());
+                    if crc_expected != crc_actual {
+                        log::warn!("crc mismatch for batch size marker");
+                        break;
+                    }
 
                     if batch_sz > usize::MAX as u64 {
                         return Err(io::Error::new(
@@ -594,40 +599,50 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
                         ));
                     }
 
-                    let crc_actual = hash_batch_len(usize::try_from(batch_sz).unwrap());
-                    if crc_expected != crc_actual {
-                        rewind!();
-                    }
+                    let wb_remaining = batch_sz as usize;
+                    let wb_recovered = buf.len() as u64;
+
+                    write_batch = dbg!(Some((
+                        Vec::with_capacity(batch_sz as usize),
+                        wb_remaining,
+                        wb_recovered,
+                    )));
 
-                    write_batch = Some(Vec::with_capacity(batch_sz as usize));
-                    wb_remaining = crc_expected as usize;
                     continue;
                 }
                 _ => {
-                    rewind!();
+                    log::warn!("invalid log message discriminant detected: {}", buf[4]);
+                    break;
                 }
             };
-            let k: [u8; K] = buf[5..K + 5].try_into().unwrap();
+            let k: [u8; K] = buf[5..5 + K].try_into().unwrap();
             let v: Option<[u8; V]> = if d {
-                Some(buf[K + 5..5 + K + V].try_into().unwrap())
+                Some(buf[5 + K..5 + K + V].try_into().unwrap())
             } else {
                 None
             };
             let crc_actual: u32 = hash(&k, &v);
 
             if crc_expected != crc_actual {
-                rewind!();
+                log::warn!(
+                    "crc mismatch for kv pair {:?}-{:?}: expected {} actual {}, torn log detected",
+                    k,
+                    v,
+                    crc_expected,
+                    crc_actual
+                );
+                break;
             }
 
-            if let Some(ref mut wb) = write_batch {
+            if let Some((mut wb, mut wb_remaining, mut wb_recovered)) = write_batch.take() {
                 wb.push((k, v));
-                wb_remaining -= 1;
-                wb_recovered += buf.len() as u64;
+                wb_remaining = wb_remaining.checked_sub(1).unwrap();
+                wb_recovered = wb_recovered.checked_add(buf.len() as u64).unwrap();
 
                 // apply the write batch all at once
                 // or never at all
                 if wb_remaining == 0 {
-                    for (k, v) in std::mem::take(wb) {
+                    for (k, v) in wb {
                         memtable.insert(k, v);
 
                         if let Some(v) = v {
@@ -637,12 +652,10 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
                         }
                     }
                     recovered += wb_recovered;
-                    wb_recovered = 0;
+                } else {
+                    write_batch = Some((wb, wb_remaining, wb_recovered));
                 }
             } else {
-                assert_eq!(0, wb_remaining);
-                assert_eq!(0, wb_recovered);
-
                 memtable.insert(k, v);
 
                 if let Some(v) = v {
@@ -655,6 +668,15 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
             }
         }
 
+        // need to back up a few bytes to chop off the torn log
+        log::debug!("recovered {} kv pairs", db.len());
+        log::debug!("rewinding log down to length {}", recovered);
+        let log_file = reader.get_mut();
+        log_file.seek(io::SeekFrom::Start(recovered))?;
+        log_file.set_len(recovered)?;
+        log_file.sync_all()?;
+        fs::File::open(path.join(SSTABLE_DIR))?.sync_all()?;
+
         let (tx, rx) = mpsc::channel();
 
         let worker_stats = Arc::new(WorkerStats {
@@ -678,7 +700,7 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 
         let lsm = Lsm {
             #[cfg(not(test))]
-            log: BufWriter::with_capacity(config.log_bufwriter_size, reader.into_inner()),
+            log: BufWriter::with_capacity(config.log_bufwriter_size as usize, reader.into_inner()),
             #[cfg(test)]
             log: tearable::Tearable::new(reader.into_inner()),
             path: path.into(),
@@ -753,9 +775,9 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
         // the zero pad is necessary because every log
         // entry must have the same length, whether
         // it's a batch size or actual kv tuple.
-        let min_tuple_sz = std::mem::size_of::<u64>().max(K + V);
-        let pad_sz = min_tuple_sz - (K + V);
-        let pad = [0; std::mem::size_of::<u64>()];
+        let tuple_sz = U64_SZ.max(K + V);
+        let pad_sz = tuple_sz - U64_SZ;
+        let pad = [0; U64_SZ];
         self.log.write_all(&pad[..pad_sz])?;
 
         for (k, v_opt) in write_batch {
@@ -791,9 +813,9 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
         // the zero pad is necessary because every log
         // entry must have the same length, whether
         // it's a batch size or actual kv tuple.
-        let min_tuple_sz = std::mem::size_of::<u64>().max(K + V);
+        let min_tuple_sz = U64_SZ.max(K + V);
         let pad_sz = min_tuple_sz - (K + V);
-        let pad = [0; std::mem::size_of::<u64>()];
+        let pad = [0; U64_SZ];
         self.log.write_all(&pad[..pad_sz])?;
 
         let logged_bytes = 4 + 1 + min_tuple_sz;
